/**********************************************************************
   Module for computation of Moments
**********************************************************************/
#include <stdio.h>
#include <errno.h>
#include <math.h>
#include "DEFS.h"
#include "MC.h"
#include "PATH.h"

#define MAX_SYMM_PATHS 64
typedef struct moments {
REAL **Mom1;
REAL **Mom2;
REAL **Mom3;
REAL **Mom4;
} MOMENTS;

static MOMENTS *Moments;
static void AccMoments(int , int , int ); /*scott added the void */
static REAL *SymmPath[MAX_SYMM_PATHS];
static int NSymmPaths3=3;
static int NSymmPaths6=6;
static int NumRates, NumTimeSteps;
static REAL *CurrPath;

InitMom(PATH *Path)
{
 int is,i,j;

 NumRates=HowManyRates(Path);
 NumTimeSteps= HowManyTimeSteps(Path);
  
 if((Moments=(MOMENTS *)
        malloc(sizeof(MOMENTS))) == NULL) {
        printf("cannot allocate buffer for MOMENTS pointers \n");
        exit(1);
  }
  if((Moments->Mom1=(REAL **)
        malloc((NumRates+1)*sizeof(REAL *))) == NULL) {
        printf("cannot allocate buffer for Mom1 pointers \n");
        exit(1);
  }
  if((Moments->Mom2=(REAL **)
        malloc((NumRates+1)*sizeof(REAL *))) == NULL) {
        printf("cannot allocate buffer for Mom2 pointers \n");
        exit(1);
  }
  if((Moments->Mom3=(REAL **)
        malloc((NumRates+1)*sizeof(REAL *))) == NULL) {
        printf("cannot allocate buffer for Mom3 pointers \n");
        exit(1);
  }
  if((Moments->Mom4=(REAL **)
        malloc((NumRates+1)*sizeof(REAL *))) == NULL) {
        printf("cannot allocate buffer for Mom4 pointers \n");
        exit(1);
  }

  for(i=0;i<=NumRates;i++) {
    if((Moments->Mom1[i]=(REAL *)
        malloc((NumTimeSteps+1)*sizeof(REAL))) == NULL) {
        printf("cannot allocate buffer for Moment1 # %d\n",i);
        exit(1);
    }
  }
    printf("finished Mom1\n");
  for(i=0;i<=NumRates;i++) {
    if((Moments->Mom2[i]=(REAL *)
        malloc((NumTimeSteps+1)*sizeof(REAL))) == NULL) {
        printf("cannot allocate buffer for Moment2 # %d\n",i);
        exit(1);
    }
  }
    printf("finished Mom2\n");
  for(i=0;i<=NumRates;i++) {
    if((Moments->Mom3[i]=(REAL *)
        malloc((NumTimeSteps+1)*sizeof(REAL))) == NULL) {
        printf("cannot allocate buffer for Moment3 # %d\n",i);
        exit(1);
    }
  }
    printf("finished Mom3\n");
  for(i=0;i<=NumRates;i++) {
    if((Moments->Mom4[i]=(REAL *)
        malloc((NumTimeSteps+1)*sizeof(REAL))) == NULL) {
        printf("cannot allocate buffer for Moment4 # %d\n",i);
        exit(1);
    }
  }
    printf("finished Mom4\n");
  for(i=1;i<=NumRates;i++)
    for(j=0;j<=NumTimeSteps;j++) {
             *(Moments->Mom1[i]+j)=0.0;
             *(Moments->Mom2[i]+j)=0.0;
             *(Moments->Mom3[i]+j)=0.0;
             *(Moments->Mom4[i]+j)=0.0;
    }

    for(is=0;is<NSymmPaths6;is++)
     if((SymmPath[is]=(REAL *)malloc((NumTimeSteps+1)*sizeof(REAL))) == NULL) {
     perror("cannot allocate buffer for SymmPaths\n");
     exit(1);
    }
    if((CurrPath=(REAL *)
        malloc((NumTimeSteps+1)*sizeof(REAL))) == NULL) {
        printf("cannot allocate buffer for CurrentRate\n");
        exit(1);
    }


return(0);
}


/********************************************************************
   This Measurement() procedure assumes that both drift and correlation 
   matrix are time independent and rate independent. It uses the local
   reflection symetry:


   ----------------*---------    t+1
                  *  *
                 * --> *
                *        *
   ------------*-----------*-    t   
                   * --> *
                     *  *
   --------------------*-----     t-1
   on odd and even time slices respectively to improve qulity of estimators.
   With this trick alone, one obtains NSymmPaths = 3 ----> Number
   of symmetry related paths with same probability.
   This and global reflection symmetry results in NSymmPaths =6 
***************************************************************************/
	
MeasMom3(PATH *Path)
    
{
  int ir,it,is;
  REAL InitValue,drift,TotDrift;

  for(ir=1;ir<=NumRates;ir++) {

     InitValue=0.0;
     drift=0.0;

     SetCurrentRateIndex(Path, ir);
     GetCurrentPath(Path, CurrPath);
     for(it=0;it<=NumTimeSteps;it++) {
         SymmPath[0][it] = *(CurrPath+it);
         SymmPath[2][it] = SymmPath[1][it] = SymmPath[0][it];
     }
     for(it=1;it<NumTimeSteps;it +=2 )
         SymmPath[1][it]=SymmPath[1][it-1]+(SymmPath[1][it+1]-SymmPath[1][it]);
     for(it=2;it<NumTimeSteps;it +=2 )
         SymmPath[2][it]=SymmPath[2][it-1]+(SymmPath[2][it+1]-SymmPath[2][it]);

     AccMoments(NSymmPaths3,ir,NumTimeSteps);

  }  /* end of loop over rates */

  return(0);
} /* End of Measurement */


MeasMom6(PATH *Path)
    
{
  int ir,it,is;
  REAL InitValue,drift,TotDrift;

  for(ir=1;ir<=NumRates;ir++) {

     InitValue=0.0;
     drift=0.0;

     SetCurrentRateIndex(Path, ir);
     GetCurrentPath(Path, CurrPath);
     for(it=0;it<=NumTimeSteps;it++) {
         SymmPath[1][it] = -(SymmPath[0][it] = *(CurrPath+it));
         SymmPath[3][it] = SymmPath[2][it] = SymmPath[0][it];
         SymmPath[5][it] = SymmPath[4][it] = SymmPath[0][it];
     }
     for(it=1;it<NumTimeSteps;it +=2 )
         SymmPath[2][it]=SymmPath[2][it-1]+(SymmPath[2][it+1]-SymmPath[2][it]);
     for(it=2;it<NumTimeSteps;it +=2 )
         SymmPath[3][it]=SymmPath[3][it-1]+(SymmPath[3][it+1]-SymmPath[3][it]);
     for(it=0;it<=NumTimeSteps;it++) {
         SymmPath[4][it] = -SymmPath[2][it];
         SymmPath[5][it] = -SymmPath[3][it];
     }

     AccMoments(NSymmPaths6,ir,NumTimeSteps);

  }  /* end of loop over rates */
  return(0);
} /* End of Measurement */



static void AccMoments(int npaths, int ir, int NumTimeSteps) /* scott put in
static void */
{
  int is,it;
  REAL rate,rate2;
  for(is=0;is<npaths;is++)
     for(it=0;it <=NumTimeSteps; it++) {
       rate = *(SymmPath[is]+it);
       rate2=rate*rate;
       *(Moments->Mom1[ir]+it) += rate/npaths;
       *(Moments->Mom2[ir]+it) += rate2/npaths;
       *(Moments->Mom3[ir]+it) += rate*rate2/npaths;
       *(Moments->Mom4[ir]+it) += rate2*rate2/npaths;
     }
}

void PrintMoments() /* scott put in the void */
{
  int ir,it;
  for(ir=1;ir<=NumRates;ir++) {
     printf("Rate # %d \n",ir);
     for(it=0;it <=NumTimeSteps; it++) {
#ifdef DOUBLE_PRECISION
       printf("Time slice # %d :  %lf %lf\n",it, 
                 (*(Moments->Mom1[ir]+it))/HowManyMeasBlocks(), 
                 (*(Moments->Mom2[ir]+it))/HowManyMeasBlocks());
#else
       printf("Time slice # %d :  %f %f\n",it,
                 (*(Moments->Mom1[ir]+it))/HowManyMeasBlocks(),
                 (*(Moments->Mom2[ir]+it))/HowManyMeasBlocks());
#endif
     }
  }
}

